package org.biojava.bio.dp;

import java.io.Serializable;

import org.biojava.bio.BioException;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceAnnotator;
import org.biojava.bio.seq.impl.ViewSequence;
import org.biojava.bio.symbol.IllegalAlphabetException;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.RangeLocation;
import org.biojava.utils.ChangeVetoException;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.bio.dist.Distribution;

/**
 * Annotates a sequence with hits to a weight-matrix.
 *
 * <p>
 * This SequenceAnnotator implementation returns a new
 * ViewSequence wrapping the underlying Sequence
 * </p>
 *
 * @author Matthew Pocock
 * @author Thomas Down
 * @author Tanya Vavouri
 */
public class WeightMatrixAnnotator implements SequenceAnnotator,
    Serializable {
  private WeightMatrix matrix;
  private double threshold;
  private final ScoreType scoreType;
  private String wmID;

  public Sequence annotate(Sequence seq) throws IllegalAlphabetException,
      BioException, ChangeVetoException {
    seq = new ViewSequence(seq);

    int cols = matrix.columns();
    Feature.Template template = new Feature.Template();
    template.source = "WeightMatrixAnnotator";
    template.type = wmID;
    Distribution[] dist = new Distribution[cols];
    Symbol[] syms = new Symbol[cols];
    for (int i = 0; i < cols; i++) {
        dist[i] = matrix.getColumn(i);
        syms[i] = seq.symbolAt(i+1);
    }
    int start = 0; int stop = cols - 1;
    for (int offset = 1;
         offset <= seq.length() - cols;
         offset++) {
        
        double score = 0;
        int j = 0;
        for (int i = start; i != stop; i = (i + 1) % cols) {
            score += Math.log(dist[j].getWeight(syms[i]));
        }
        start = (start + 1) % cols;
        stop = (stop + 1) % cols;
        syms[stop] = seq.symbolAt(offset+cols);
        
        double q = Math.exp(score);
        if (q >= threshold) {
            template.location = new RangeLocation(offset, offset + cols - 1);
            SimpleAnnotation ann = new SimpleAnnotation();
            ann.setProperty("score", new Double(q));
            ann.setProperty("weightMatrix", matrix);
            template.annotation = ann;
            seq.createFeature(template);
        }
    }
    return seq;
  }

  /**
   * Scores the SymbolList from symbol start to symbol (start+columns) with a
   * weight matrix.
   *
   * @param matrix  the weight matrix used to evaluate the sequences
   * @param symList the SymbolList to assess
   * @param start   the index of the first symbol in the window to evaluate
   * @return  the log probability or likelyhood of this weight matrix
   *          having generated symbols start to (start + columns) of symList
   */
  public static double scoreWeightMatrix(
          WeightMatrix matrix, SymbolList symList, int start)
          throws IllegalSymbolException {
    double score = 0;
    int cols = matrix.columns();

    for (int c = 0; c < cols; c++) {
      score += Math.log(
              matrix.getColumn(c).getWeight(symList.symbolAt(c + start)));
    }

    return score;
  }

  /**
   * Scores the SymbolList from symbol start to symbol (start+columns) with a
   * weight matrix using a particular ScoreType.
   *
   * <p>
   * This method allows you to use score types such as ScoreType.ODDS. The other
   * scoreWeightMatrix methods gives a result similar or identical to
   * ScoreType.PROBABILITY.
   * </p>
   *
   * @param matrix  the weight matrix used to evaluate the sequences
   * @param symList the SymbolList to assess
   * @param scoreType the score type to apply
   * @param start   the index of the first symbol in the window to evaluate
   * @return  the sum of log scores of this weight matrix
   *          having generated symbols start to (start + columns) of symList
   * @since 1.4
   */
  public static double scoreWeightMatrix(
          WeightMatrix matrix,
          SymbolList symList,
          ScoreType scoreType,
          int start)
          throws IllegalSymbolException {
    double score = 0;
    int cols = matrix.columns();

    for (int c = 0; c < cols; c++) {
      score += Math.log(scoreType.calculateScore(
              matrix.getColumn(c), symList.symbolAt(c + start)));
    }

    return score;
  }


  /**
   * Create a new annotator that uses the PROBABILITY score type and an ID
   for the weight matrix.
   *
   * @param wm        the weight matrix
   * @param threshold the threshold
   * @param wmID the weight matrix ID
   */
  public WeightMatrixAnnotator(WeightMatrix wm, ScoreType scoreType,
                               double threshold, String wmID) {
    this.matrix = wm;
    this.threshold = threshold;
    this.scoreType = scoreType;
    this.wmID = wmID;
  }

  /**
   * Create a new annotator that uses PROBABILITY score type.
   *
   * @param wm a <code>WeightMatrix</code> value
   * @param threshold a <code>double</code> value
   */
  public WeightMatrixAnnotator(WeightMatrix wm, double threshold) {
    this(wm, ScoreType.PROBABILITY, threshold, "hit");
  }

  /**
   * Create a new annotator that uses a specific score type.
   *
   * @param wm        the weigth matrix
   * @param scoreType the score type
   * @param threshold the threshold
   * @since 1.4
   */
  public WeightMatrixAnnotator(WeightMatrix wm, ScoreType scoreType,
                               double threshold) {
    this.matrix = wm;
    this.scoreType = scoreType;
    this.threshold = threshold;
    this.wmID = "hit";
  }

  /**
   * Get the value of the weight matrix id.
   * @return value of the weight matrix id.
   */
  public String getWeightMatrixID() {
    return wmID;
  }

  /**
   * Set the weight matrix id.
   * @param id  Value to assign to the weight matrix id.
   */
  public void setWeightMatrixID(String id) {
    this.wmID = id;
  }

}
